'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var data = require('./data.cjs.js');
var cssVar = require('./utils/css-var.cjs.js');
var noop = require('../../utils/noop.cjs.js');
var time = require('./utils/time.cjs.js');
var transforms = require('./utils/transforms.cjs.js');
var easing = require('./utils/easing.cjs.js');
var featureDetection = require('./utils/feature-detection.cjs.js');
var apply = require('./utils/apply.cjs.js');
var NumberAnimation = require('../js/NumberAnimation.cjs.js');
var keyframes = require('./utils/keyframes.cjs.js');
var style = require('./style.cjs.js');
var defaults = require('./utils/defaults.cjs.js');
var getStyleName = require('./utils/get-style-name.cjs.js');
var isNumber = require('../../utils/is-number.cjs.js');
var stopAnimation = require('./utils/stop-animation.cjs.js');

function animateStyle(element, key, keyframesDefinition, options = {}) {
    let animation;
    let { duration = defaults.defaults.duration, delay = defaults.defaults.delay, endDelay = defaults.defaults.endDelay, repeat = defaults.defaults.repeat, easing: easing$1 = defaults.defaults.easing, direction, offset, allowWebkitAcceleration = false, } = options;
    const data$1 = data.getAnimationData(element);
    let canAnimateNatively = featureDetection.supports.waapi();
    let render = noop.noop;
    const valueIsTransform = transforms.isTransform(key);
    /**
     * If this is an individual transform, we need to map its
     * key to a CSS variable and update the element's transform style
     */
    valueIsTransform && transforms.addTransformToElement(element, key);
    const name = getStyleName.getStyleName(key);
    /**
     * Get definition of value, this will be used to convert numerical
     * keyframes into the default value type.
     */
    const definition = transforms.transformDefinitions.get(name);
    /**
     * Stop the current animation, if any. Because this will trigger
     * commitStyles (DOM writes) and we might later trigger DOM reads,
     * this is fired now and we return a factory function to create
     * the actual animation that can get called in batch,
     */
    stopAnimation.stopAnimation(data$1.animations[name]);
    /**
     * Batchable factory function containing all DOM reads.
     */
    return () => {
        const readInitialValue = () => { var _a, _b; return (_b = (_a = style.style.get(element, name)) !== null && _a !== void 0 ? _a : definition === null || definition === void 0 ? void 0 : definition.initialValue) !== null && _b !== void 0 ? _b : 0; };
        /**
         * Replace null values with the previous keyframe value, or read
         * it from the DOM if it's the first keyframe.
         */
        let keyframes$1 = keyframes.hydrateKeyframes(keyframes.keyframesList(keyframesDefinition), readInitialValue);
        if (easing.isCustomEasing(easing$1)) {
            const custom = easing$1.createAnimation(keyframes$1, readInitialValue, valueIsTransform, name, data$1);
            easing$1 = custom.easing;
            if (custom.keyframes !== undefined)
                keyframes$1 = custom.keyframes;
            if (custom.duration !== undefined)
                duration = custom.duration;
        }
        /**
         * If this is a CSS variable we need to register it with the browser
         * before it can be animated natively. We also set it with setProperty
         * rather than directly onto the element.style object.
         */
        if (cssVar.isCssVar(name)) {
            render = apply.cssVariableRenderer(element, name);
            if (featureDetection.supports.cssRegisterProperty()) {
                cssVar.registerCssVariable(name);
            }
            else {
                canAnimateNatively = false;
            }
        }
        else {
            render = apply.styleRenderer(element, name);
        }
        /**
         * If we can animate this value with WAAPI, do so. Currently this only
         * feature detects CSS.registerProperty but could check WAAPI too.
         */
        if (canAnimateNatively) {
            /**
             * Convert numbers to default value types. Currently this only supports
             * transforms but it could also support other value types.
             */
            if (definition) {
                keyframes$1 = keyframes$1.map((value) => isNumber.isNumber(value) ? definition.toDefaultUnit(value) : value);
            }
            /**
             * If this browser doesn't support partial/implicit keyframes we need to
             * explicitly provide one.
             */
            if (!featureDetection.supports.partialKeyframes() && keyframes$1.length === 1) {
                keyframes$1.unshift(readInitialValue());
            }
            const animationOptions = {
                delay: time.ms(delay),
                duration: time.ms(duration),
                endDelay: time.ms(endDelay),
                easing: !easing.isEasingList(easing$1) ? easing.convertEasing(easing$1) : undefined,
                direction,
                iterations: repeat + 1,
                fill: "both",
            };
            animation = element.animate({
                [name]: keyframes$1,
                offset,
                easing: easing.isEasingList(easing$1) ? easing$1.map(easing.convertEasing) : undefined,
            }, animationOptions);
            /**
             * Polyfill finished Promise in browsers that don't support it
             */
            if (!animation.finished) {
                animation.finished = new Promise((resolve, reject) => {
                    animation.onfinish = resolve;
                    animation.oncancel = reject;
                });
            }
            const target = keyframes$1[keyframes$1.length - 1];
            animation.finished
                .then(() => {
                // Apply styles to target
                render(target);
                // Ensure fill modes don't persist
                animation.cancel();
            })
                .catch(noop.noop);
            /**
             * This forces Webkit to run animations on the main thread by exploiting
             * this condition:
             * https://trac.webkit.org/browser/webkit/trunk/Source/WebCore/platform/graphics/ca/GraphicsLayerCA.cpp?rev=281238#L1099
             *
             * This fixes Webkit's timing bugs, like accelerated animations falling
             * out of sync with main thread animations and massive delays in starting
             * accelerated animations in WKWebView.
             */
            if (!allowWebkitAcceleration)
                animation.playbackRate = 1.000001;
            /**
             * If we can't animate the value natively then we can fallback to the numbers-only
             * polyfill for transforms. All keyframes must be numerical.
             */
        }
        else if (valueIsTransform && keyframes$1.every(isNumber.isNumber)) {
            /**
             * If we only have a single keyframe, we need to create an initial keyframe by reading
             * the current value from the DOM.
             */
            if (keyframes$1.length === 1) {
                keyframes$1.unshift(parseFloat(readInitialValue()));
            }
            if (definition) {
                const applyStyle = render;
                render = (v) => applyStyle(definition.toDefaultUnit(v));
            }
            animation = new NumberAnimation.NumberAnimation(render, keyframes$1, Object.assign(Object.assign({}, options), { duration,
                easing: easing$1 }));
        }
        else {
            const target = keyframes$1[keyframes$1.length - 1];
            render(definition && isNumber.isNumber(target)
                ? definition.toDefaultUnit(target)
                : target);
        }
        data$1.animations[name] = animation;
        /**
         * When an animation finishes, delete the reference to the previous animation.
         */
        animation === null || animation === void 0 ? void 0 : animation.finished.then(() => {
            data$1.animations[name] = undefined;
            data$1.generators[name] = undefined;
            data$1.prevGeneratorState[name] = undefined;
        }).catch(noop.noop);
        return animation;
    };
}

exports.animateStyle = animateStyle;
