'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var defaults = require('../dom/utils/defaults.cjs.js');
var easing = require('../dom/utils/easing.cjs.js');
var getFunction = require('./easing/utils/get-function.cjs.js');
var interpolate = require('./utils/interpolate.cjs.js');

class NumberAnimation {
    constructor(output, keyframes = [0, 1], { easing: easing$1 = defaults.defaults.easing, duration = defaults.defaults.duration, delay = defaults.defaults.delay, endDelay = defaults.defaults.endDelay, repeat = defaults.defaults.repeat, offset, direction = "normal", } = {}) {
        this.startTime = 0;
        this.rate = 1;
        this.t = 0;
        this.cancelTimestamp = 0;
        this.playState = "idle";
        this.finished = new Promise((resolve, reject) => {
            this.resolve = resolve;
            this.reject = reject;
        });
        const totalDuration = duration * (repeat + 1);
        /**
         * We don't currently support custom easing (spring, glide etc) in NumberAnimation
         * (although this is completely possible), so this will have been hydrated by
         * animateStyle.
         */
        if (easing.isCustomEasing(easing$1))
            easing$1 = "ease";
        const interpolate$1 = interpolate.slowInterpolateNumbers(keyframes, offset, easing.isEasingList(easing$1)
            ? easing$1.map(getFunction.getEasingFunction)
            : getFunction.getEasingFunction(easing$1));
        this.tick = (timestamp) => {
            if (this.pauseTime)
                timestamp = this.pauseTime;
            let t = (timestamp - this.startTime) * this.rate;
            this.t = t;
            // Convert to seconds
            t /= 1000;
            // Rebase on delay
            t = Math.max(t - delay, 0);
            /**
             * If this animation has finished, set the current time
             * to the total duration.
             */
            if (this.playState === "finished")
                t = totalDuration;
            /**
             * Get the current progress (0-1) of the animation. If t is >
             * than duration we'll get values like 2.5 (midway through the
             * third iteration)
             */
            const progress = t / duration;
            // TODO progress += iterationStart
            /**
             * Get the current iteration (0 indexed). For instance the floor of
             * 2.5 is 2.
             */
            let currentIteration = Math.floor(progress);
            /**
             * Get the current progress of the iteration by taking the remainder
             * so 2.5 is 0.5 through iteration 2
             */
            let iterationProgress = progress % 1.0;
            if (!iterationProgress && progress >= 1) {
                iterationProgress = 1;
            }
            /**
             * If iteration progress is 1 we count that as the end
             * of the previous iteration.
             */
            iterationProgress === 1 && currentIteration--;
            /**
             * Reverse progress if we're not running in "normal" direction
             */
            const iterationIsOdd = currentIteration % 2;
            if (direction === "reverse" ||
                (direction === "alternate" && iterationIsOdd) ||
                (direction === "alternate-reverse" && !iterationIsOdd)) {
                iterationProgress = 1 - iterationProgress;
            }
            const latest = interpolate$1(t >= totalDuration ? 1 : Math.min(iterationProgress, 1));
            output(latest);
            const isAnimationFinished = this.playState === "finished" || t >= totalDuration + endDelay;
            if (isAnimationFinished) {
                this.playState = "finished";
                this.resolve(latest);
            }
            else if (this.playState !== "idle") {
                requestAnimationFrame(this.tick);
            }
        };
        this.play();
    }
    play() {
        const now = performance.now();
        this.playState = "running";
        if (this.pauseTime) {
            this.startTime = now - (this.pauseTime - this.startTime);
        }
        else if (!this.startTime) {
            this.startTime = now;
        }
        this.pauseTime = undefined;
        requestAnimationFrame(this.tick);
    }
    pause() {
        this.playState = "paused";
        this.pauseTime = performance.now();
    }
    finish() {
        this.playState = "finished";
        this.tick(0);
    }
    cancel() {
        this.playState = "idle";
        this.tick(this.cancelTimestamp);
        this.reject(false);
    }
    reverse() {
        this.rate *= -1;
    }
    commitStyles() {
        this.cancelTimestamp = performance.now();
    }
    get currentTime() {
        return this.t;
    }
    set currentTime(t) {
        if (this.pauseTime || this.rate === 0) {
            this.pauseTime = t;
        }
        else {
            this.startTime = performance.now() - t / this.rate;
        }
    }
    get playbackRate() {
        return this.rate;
    }
    set playbackRate(rate) {
        this.rate = rate;
    }
}

exports.NumberAnimation = NumberAnimation;
